'' =================================================================================================
''
''   File....... hc-8+_vsa_dmx_player.spin
''   Purpose.... 
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2011-2012 McPhalen
''               -- see below for terms of use
''   E-mail..... 
''   Started.... 
''   Updated.... 24 JUN 2012
''
'' =================================================================================================

{

  Plays show created in VSA
  -- output is to DMX (T+/T-) pins
  -- file must be exported from VSA in binary mode
  -- requires EFX-TEK uSD accessory for HC-8+
  -- all OPTION switches must be set to the OFF position for SD card to work

}


con

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' 5MHz crystal

  CLK_FREQ = ((_clkmode - xtal1) >> 6) * _xinfreq
  MS_001   = CLK_FREQ / 1_000
  US_001   = CLK_FREQ / 1_000_000
  

con

  PCB_REV = "C"

  RX1     = 31
  TX1     = 30
  SDA     = 29
  SCL     = 28

  OUT_EN  = 27      
  
  SIO     = 26
  
  RX2     = 23
  RX2_EN  = 22
  TX2_EN  = 21
  TX2     = 20

  R_LED   = 18
  G_LED   = 17

  OUT7    = 15
  OUT6    = 14   
  OUT5    = 13 
  OUT4    = 12
  OUT3    = 11
  OUT2    = 10   
  OUT1    =  9 
  OUT0    =  8
          
  OPT_BR  =  7
  OPT_A1  =  6   
  OPT_A0  =  5
  OPT_SM  =  4

  DMX_A8  =  3

  DO_165  =  2
  CLK_165 =  1 
  LD_165  =  0


  ' for EFX-TEK uSD adapter module
  ' -- must open all option switches when installed

  SD_CS   =  OPT_BR                                       
  SD_DI   =  OPT_A1  
  SD_CLK  =  OPT_A0  
  SD_DO   =  OPT_SM


con

  CHANNELS = 8                                                  ' channels in VSA show
  FR_SIZE  = CHANNELS << 2                                      ' bytes per frame
  
  FPS      = 30                                                 ' frame rate of VSA show
  

con

   #0, OFF, RED, GRN, YEL                                       ' for r/g led

   #0, NOCON, RX, TX, RXTX                                      ' for rs-485

   #0, LSBFIRST, MSBFIRST                                       ' for 74x165 shift-register 
  

con

   #1, HOME, GOTOXY, #8, BKSP, TAB, LF, CLREOL, CLRDN, CR       ' PST formmatting control
  #14, GOTOX, GOTOY, CLS


obj

  sd   : "fsrw"                                                 ' SD card IO
  dmx  : "efx_dmxout"                                           ' DMX output
  leds : "efx_pwm8"                                             ' LED dimming ***
  term : "fullduplexserial64"                                   ' terminal IO
       

var

  long  framebuf[FR_SIZE]                                       ' buffer for raw frame data


dat

Banner                  byte    "=========================", CR
                        byte    " EFX-TEK VSA Show Player ", CR
                        byte    "=========================", CR
                        byte    0

MsgFail                 byte    "-- failed", CR, 0
MsgSuccess              byte    "-- success", CR, 0                           

Msg1                    byte    CR, "Mounting SD card", CR, 0
Msg2                    byte    CR, "Waiting for selection", CR, 0
Msg3                    byte    "-- playing: ", 0
Msg3f                   byte    "-- file error", CR, 0
Msg3s                   byte    "-- done; frame count = ", 0                  


ShowFile                byte    "SHOWxx.BIN", 0                 ' show filename


pub main | check, show

  setup_io                                                      ' configure HC-8 io pins
  dmx.start(TX2, TX2_EN, 0, 32, 11_1)                           ' for 32-ch show
  leds.start(8, OUT0)                                           ' start led driver ***                    ' 
  term.start(RX1, TX1, %0000, 115_200)                          ' start terminal
  pause(1)

  term.tx(CLS)
  term.str(@Banner)                                             ' show program title
  
  term.str(@Msg1)
  check := \sd.mount_explicit(SD_DO, SD_CLK, SD_DI, SD_CS)      ' mount SD card
  if (check == 0)
    term.str(@MsgSuccess)
  else
    term.str(@MsgFail)
    repeat                                                      ' flash red if no SD card
      set_rgled(RED)
      pause(100)
      set_rgled(OFF)
      pause(100)

  ' main program loop 

  repeat                                                        
    term.str(@Msg2)
    repeat
      show := rd_matrix                                         ' wait for input
    until (show > 0)                                           ' valid = 0..7

    byte[@ShowFile][4] := (show  / 10) + "0"                    ' add number to filename
    byte[@ShowFile][5] := (show // 10) + "0"

    term.str(@Msg3)                                             ' show selection           
    term.str(@ShowFile)
    term.tx(CR)

    check := play_show(@ShowFile, FPS)                          ' play the selected show
    if (check > 0)                                              ' good play?
      term.str(@Msg3s)
      term.dec(check)                                           ' show frames played
      term.tx(CR)
    else                                                        ' file error?
      term.str(@Msg3f)
      set_rgled(RED)                                            ' solid red
      pause(1000)                                               ' at least one second
      repeat                                                    ' force button release
        check := rd_matrix                                      ' rescan
      until (check == 0)
      set_rgled(OFF)                                            ' ready for another attempt 


con

  { ========================= }
  {                           }
  {  M A T R I X   I N P U T  } 
  {                           }
  { ========================= }


pub rd_matrix | btns

'' Reads 7x2 button matrix on IN7..IN0
'' -- returns 0 (no button) to 14 (button #)
'' -- requires 7x2 matrix with steering diodes
''    * IN0 is shift input

  btns := ttl_inputs(true)                                      ' read ttl inputs

  if (btns & %0000_0001)                                        ' shift bit active?
    btns := (btns & %1111_1110) << 7                            ' remove shift, adjust

  if (bit_count(btns) == 1)                                     ' only one button pressed?
    return bit_pos(btns, 0)                                     '  yes, return button #
  else
    return 0                                                    '  none or bad input
     

con

  { =========================== }
  {                             }
  {  P L A Y   V S A   S H O W  } 
  {                             }
  { =========================== }


pub has_file(fnpntr) | check

'' Checks for presence of file on SD card
'' -- fnpntr is pointer to z-string holding filename

  check := \sd.popen(fnpntr, "r")                               ' attempt to open

  if (check == 0)
    return true
  else
    return false  
  

pub play_show(fnpntr, rate) | check, frames, t, ch

'' Open VSA binary show file and play to end
'' -- fnpntr pointer to file name (z-string)
'' -- rate is frames per second, usually 30

  if (has_file(fnpntr) == false)                                ' if no file
    return -1                                                   '  abort with error

  set_rgled(GRN)                                                ' show playback
  frames := 0                                                   ' clear frame count 
    
  t := cnt                                                      ' sync timer
  repeat
    check := \sd.pread(@framebuf, FR_SIZE)                      ' read one frame
    if (check <> FR_SIZE)                                         ' if end-of-file
      quit                                                      ' abort
    else
      frames += 1                                               ' increment frame count
      repeat ch from 0 to CHANNELS-1                            ' copy frame data to dmx
        if (framebuf[ch] => 0)                                  ' if changed
          dmx.set(ch+1, framebuf[ch])                           '  write to dmx
          leds.set(ch, framebuf[ch])                            '  write to local leds ***
      waitcnt(t += clkfreq / rate)                              ' let frame time expire

  set_rgled(OFF)
      
  \sd.pclose
  return frames                                                 ' show played
  
      
con

  { ================================= }
  {                                   }
  {  S U P P O R T   R O U T I N E S  } 
  {                                   }
  { ================================= }


var

  long  lastscan                                                ' ttl/DMX input scan


pub setup_io

'' Configure for basic IO control

  outa[OUT7..OUT0] := %0000_0000                                ' clear outputs
  dira[OUT7..OUT0] := %1111_1111                                ' output mode                                  
  low(OUT_EN)                                                   ' enable 74x245 

  high(LD_165)                                                  ' setup x165 io pins
  low(CLK_165)
  input(DO_165)
  
  set_rs485(NOCON)                                              ' RS-485 off


pub pause(ms) | t

'' Delay program in milliseconds

  if (ms < 1)                                                   ' delay must be at least 1
    return
  else
    t := cnt - 1792                                             ' sync with system counter
    repeat ms                                                   ' run delay
      waitcnt(t += MS_001)


pub high(pin)

'' Makes pin output and high

  outa[pin] := 1
  dira[pin] := 1


pub low(pin)

'' Makes pin output and low

  outa[pin] := 0
  dira[pin] := 1


pub toggle(pin)

'' Toggles output pin

  !outa[pin]                                                    ' invert state
  dira[pin] := 1  
  

pub input(pin)

  dira[pin] := 0

  return ina[pin]


pub ttl_pin | btns

'' Reads TTL header for 1-of-8 input
'' -- returns -1 (no input)
'' -- returns 0 to 7 for input on IN0 to IN7
''    * only one input allowed

  btns := ttl_inputs(true)                                      ' read ttl inputs

  if (bit_count(btns) == 1)                                     ' only one button pressed?
    return bit_pos(btns, 0)                                     '  yes, return button #
  else
    return -1                                                   '  none or bad input


pub ttl_inputs(rescan) 

'' Return TTL inputs status
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in(16, MSBFIRST) | (ina[DMX_A8] << 16)     ' read in0..in7, dmx addresss
   
  return (lastscan & $FF)                                       ' return TTL input bits


pub dmx_address(rescan) 

'' Return DMX address switch setting
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in(16, MSBFIRST) | (ina[DMX_A8] << 16)     ' read in0..in7, dmx addresss

  return (lastscan >> 8) & $01FF                                ' return DMX address bits
 

pub x165_in(bits, mode) | tmp165

'' Input value from 74x165(s)      
                                                                                          
  outa[LD_165] := 0                                             ' blip Shift/Load line    
  outa[LD_165] := 1                                                                         
                                                                                          
  tmp165 := 0                                                   ' clear workspace
  bits := 1 #> bits <# 32                                       ' limit to legal value         
  repeat bits                                                   ' read n bits         
    tmp165 := (tmp165 << 1) | ina[DO_165]                       ' get new bit             
    outa[CLK_165] := 1                                          ' blip clock              
    outa[CLK_165] := 0  

  if (mode == LSBFIRST)                                         ' LSBFIRST result?
    tmp165 ><= bits                                             ' reverse bits
   
  return tmp165
  

pub option_inputs

'' Returns state of option switch bits (%0000..%1111, all off to all on)

  dira[OPT_BR..OPT_SM] := %0000                                 ' force to input mode

  return !ina[OPT_BR..OPT_SM]                                   ' return state (1 = on)


pub set_rgled(state)

'' Sets R/G LED
'' -- direct control only
'' -- for red/green/yellow use "efx_rgled" driver and methods

  case state
    OFF : outa[R_LED..G_LED] := %00
    RED : outa[R_LED..G_LED] := %10
    GRN : outa[R_LED..G_LED] := %01
    YEL : outa[R_LED..G_LED] := %00                             ' yellow requires PASM driver

  dira[R_LED..G_LED] := %11
      

pub set_rs485(state)

'' Sets RS-485 RX and TX enable inputs

  case state
    NOCON : outa[RX2_EN..TX2_EN] := %10
    RX    : outa[RX2_EN..TX2_EN] := %00 
    TX    : outa[RX2_EN..TX2_EN] := %11   
    RXTX  : outa[RX2_EN..TX2_EN] := %01

  dira[RX2_EN..TX2_EN] := %11


pub bit_count(value) | bc

'' Returns # of bits set (1) in value

  bc := 0                                                       ' clear count
  repeat 32                                                     ' test all bits
    bc += (value & %1)                                          ' add bit value to count
    value >>= 1                                                 ' next bit

  return bc


pub bit_pos(value, mode)

'' Returns position of 1st bit set (0..31)
'' -- mode 0 to scan from lsb, mode 1 to scan from msb 
'' -- -1 = no bits set

  if (value == 0)                                               ' if no bits                                   
    return -1                                                   '  return -1

  else
    if (mode == 0)                                              ' check from LSB
      value ><= 32                                              '  flip for >|
      return (32 - >|value)
    else
      return (>|value - 1)


dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}  